@@ -1,8 +1,26 @@
+from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
+
+import matplotlib.pyplot as plt
 import numpy as np
-import theano.tensor as tt
+import pandas as pd
+from matplotlib import cm
+from matplotlib.axes import Axes
+from matplotlib.colors import Colormap
 from scipy.special import logsumexp
 
-vsearchsorted = np.vectorize(np.searchsorted, otypes=[np.int], signature="(n),()->()")
+try:  # pragma: no cover
+    import aesara.tensor as at
+    from aesara.tensor.extra_ops import broadcast_shape
+    from aesara.tensor.extra_ops import broadcast_to as at_broadcast_to
+    from aesara.tensor.var import TensorVariable
+except ImportError:  # pragma: no cover
+    import theano.tensor as at
+    from theano.tensor.extra_ops import broadcast_shape
+    from theano.tensor.extra_ops import broadcast_to as at_broadcast_to
+    from theano.tensor.var import TensorVariable
+
+
+vsearchsorted = np.vectorize(np.searchsorted, otypes=[int], signature="(n),()->()")
 
 
 def compute_steady_state(P):
@@ -20,8 +38,8 @@ def compute_steady_state(P):
 
     P = P[0]
     N_states = P.shape[-1]
-    Lam = (tt.eye(N_states) - P + tt.ones((N_states, N_states))).T
-    u = tt.slinalg.solve(Lam, tt.ones((N_states,)))
+    Lam = (at.eye(N_states) - P + at.ones((N_states, N_states))).T
+    u = at.slinalg.solve(Lam, at.ones((N_states,)))
     return u
 
 
@@ -71,15 +89,15 @@ def compute_trans_freqs(states, N_states, counts_only=False):
 
 def tt_logsumexp(x, axis=None, keepdims=False):
     """Construct a Theano graph for a log-sum-exp calculation."""
-    x_max_ = tt.max(x, axis=axis, keepdims=True)
+    x_max_ = at.max(x, axis=axis, keepdims=True)
 
     if x_max_.ndim > 0:
-        x_max_ = tt.set_subtensor(x_max_[tt.isinf(x_max_)], 0.0)
-    elif tt.isinf(x_max_):
-        x_max_ = tt.as_tensor(0.0)
+        x_max_ = at.set_subtensor(x_max_[at.isinf(x_max_)], 0.0)
+    elif at.isinf(x_max_):
+        x_max_ = at.as_tensor(0.0)
 
-    res = tt.sum(tt.exp(x - x_max_), axis=axis, keepdims=keepdims)
-    res = tt.log(res)
+    res = at.sum(at.exp(x - x_max_), axis=axis, keepdims=keepdims)
+    res = at.log(res)
 
     if not keepdims:
         # SciPy uses the `axis` keyword here, but Theano doesn't support that.
@@ -102,10 +120,12 @@ def tt_logdotexp(A, b):
     The result is more or less equivalent to `tt.log(tt.exp(A).dot(tt.exp(b)))`
 
     """
-    A_bcast = A.dimshuffle(list(range(A.ndim)) + ["x"])
+    pattern: List[Any] = list(range(A.ndim))
+    A_bcast = A.dimshuffle(pattern + ["x"])
 
     sqz = False
-    shape_b = ["x"] + list(range(b.ndim))
+    pattern_b: List[Any] = list(range(b.ndim))
+    shape_b = ["x"] + pattern_b
     if len(shape_b) < 3:
         shape_b += ["x"]
         sqz = True
@@ -149,7 +169,7 @@ def tt_expand_dims(x, dims):
         Position in the expanded axes where the new axis (or axes) is placed.
 
     """
-    dim_range = list(range(x.ndim))
+    dim_range: List[Any] = list(range(x.ndim))
     for d in sorted(np.atleast_1d(dims), reverse=True):
         offset = 0 if d >= 0 else len(dim_range) + 1
         dim_range.insert(d + offset, "x")
@@ -157,50 +177,17 @@ def tt_expand_dims(x, dims):
     return x.dimshuffle(dim_range)
 
 
-def tt_broadcast_arrays(*args):
+def tt_broadcast_arrays(*args: TensorVariable):
     """Broadcast any number of arrays against each other.
 
-    This is a Theano emulation of `numpy.broadcast_arrays`.  It does *not* use
-    memory views, and--as a result--it will not be nearly as efficient as the
-    NumPy version.
-
     Parameters
     ----------
     `*args` : array_likes
         The arrays to broadcast.
 
     """
-    p = max(a.ndim for a in args)
-
-    args = [tt.shape_padleft(a, n_ones=p - a.ndim) if a.ndim < p else a for a in args]
-
-    bcast_shape = [None] * p
-    for i in range(p - 1, -1, -1):
-        non_bcast_args = [tuple(a.shape)[i] for a in args if not a.broadcastable[i]]
-        bcast_shape[i] = tt.max([1] + non_bcast_args)
-
-    # TODO: This could be very costly?
-    return [a * tt.ones(bcast_shape) for a in args]
-
-
-def broadcast_to(x, shape):
-    """Broadcast an array to a new shape.
-
-    This implementation will use NumPy when an `ndarray` is given and an
-    inefficient Theano variant otherwise.
-
-    Parameters
-    ----------
-    x : array_like
-        The array to broadcast.
-    shape : tuple
-        The shape of the desired array.
-    """
-    if isinstance(x, np.ndarray):
-        return np.broadcast_to(x, shape)  # pragma: no cover
-    else:
-        # TODO: This could be very costly?
-        return x * tt.ones(shape)
+    bcast_shape = broadcast_shape(*args)
+    return tuple(at_broadcast_to(a, bcast_shape) for a in args)
 
 
 def multilogit_inv(ys):
@@ -223,7 +210,7 @@ def multilogit_inv(ys):
         lib = np
         lib_logsumexp = logsumexp
     else:
-        lib = tt
+        lib = at
         lib_logsumexp = tt_logsumexp
 
     # exp_ys = lib.exp(ys)
@@ -236,16 +223,15 @@ def multilogit_inv(ys):
 
 
 def plot_split_timeseries(
-    data,
-    split_freq="W",
-    split_max=5,
-    twin_column_name=None,
-    twin_plot_kwargs=None,
-    figsize=(15, 15),
-    title=None,
-    drawstyle="steps-pre",
-    linewidth=0.5,
-    plot_fn=None,
+    data: pd.DataFrame,
+    axes_split_data: Optional[List] = None,
+    split_freq: str = "W",
+    split_max: int = 5,
+    twin_column_name: Optional[str] = None,
+    twin_plot_kwargs: Optional[Dict[str, Any]] = None,
+    figsize: Tuple[int, ...] = (15, 15),
+    title: Optional[str] = None,
+    plot_fn: Optional[Callable[..., Any]] = None,
     **plot_kwds
 ):  # pragma: no cover
     """Plot long timeseries by splitting them across multiple rows using a given time frequency.
@@ -256,6 +242,8 @@ def plot_split_timeseries(
     ----------
     data: DataFrame
         The timeseries to be plotted.
+    axes_split_data: List (optional)
+        The result from a previous call of `plot_split_timeseries`.
     split_freq: str
         A Pandas time frequency string by which the series is split.
     split_max: int
@@ -275,16 +263,16 @@ def plot_split_timeseries(
     axes : list of axes
         The generated plot axes.
     """  # noqa: E501
-    import matplotlib.dates as mdates
-    import matplotlib.pyplot as plt
     import matplotlib.transforms as mtrans
-    import pandas as pd
+    from matplotlib.dates import AutoDateFormatter, AutoDateLocator
 
     if plot_fn is None:
 
-        def plot_fn(ax, data, **kwargs):
+        def plot_fn_(ax, data, **kwargs):
             return ax.plot(data, **kwargs)
 
+        plot_fn = plot_fn_
+
     data = pd.DataFrame(data)
 
     if twin_column_name and len(data.columns) < 2:
@@ -294,19 +282,31 @@ def plot_split_timeseries(
 
     split_offset = pd.tseries.frequencies.to_offset(split_freq)
 
-    grouper = pd.Grouper(freq=split_offset.freqstr, closed="left")
-    obs_splits = [y_split for n, y_split in data.groupby(grouper)]
+    if axes_split_data is None:
+        grouper = pd.Grouper(freq=split_offset.freqstr, closed="left")
+        obs_splits = [y_split for n, y_split in data.groupby(grouper)]
 
-    if split_max:
-        obs_splits = obs_splits[:split_max]
+        if split_max:
+            obs_splits = obs_splits[:split_max]
 
-    n_partitions = len(obs_splits)
+        n_partitions = len(obs_splits)
 
-    fig, axes = plt.subplots(
-        nrows=n_partitions, sharey=True, sharex=False, figsize=figsize
-    )
+        fig, axes = plt.subplots(
+            nrows=n_partitions, sharey=True, sharex=False, figsize=figsize
+        )
 
-    major_offset = mtrans.ScaledTranslation(0, -10 / 72.0, fig.dpi_scale_trans)
+        major_offset = mtrans.ScaledTranslation(0, -10 / 72.0, fig.dpi_scale_trans)
+
+    else:
+        # access split data
+        if isinstance(axes_split_data[0][0], Axes):
+            axes, split_indices = zip(*axes_split_data)
+        else:
+            axes_, splits_ = zip(*axes_split_data)
+            (axes, _), (splits, _) = zip(*axes_), zip(*splits_)
+            split_indices = [split.index for split in splits]
+        obs_splits = [data.loc[split_idx, :] for split_idx in split_indices]
+        major_offset = None
 
     axes[0].set_title(title)
 
@@ -318,16 +318,18 @@ def plot_split_timeseries(
             alt_data = split_data[twin_column_name].to_frame()
             split_data = split_data.drop(columns=[twin_column_name])
 
-        plot_fn(ax, split_data, drawstyle=drawstyle, linewidth=linewidth, **plot_kwds)
+        plot_fn(ax, split_data, **plot_kwds)
 
-        ax.xaxis.set_minor_locator(mdates.HourLocator(byhour=range(0, 23, 3)))
-        ax.xaxis.set_minor_formatter(mdates.DateFormatter("%H"))
-        ax.xaxis.set_major_locator(mdates.WeekdayLocator(byweekday=range(0, 7, 1)))
-        ax.xaxis.set_major_formatter(mdates.DateFormatter("%Y-%m-%d %a"))
+        if major_offset is not None:
+            locator = AutoDateLocator()
+            formatter = AutoDateFormatter(locator)
 
-        # Shift the major tick labels down
-        for xlabel in ax.xaxis.get_majorticklabels():
-            xlabel.set_transform(xlabel.get_transform() + major_offset)
+            ax.xaxis.set_major_locator(locator)
+            ax.xaxis.set_major_formatter(formatter)
+
+            # Shift the major tick labels down
+            for xlabel in ax.xaxis.get_majorticklabels():
+                xlabel.set_transform(xlabel.get_transform() + major_offset)
 
         legend_lines, legend_labels = ax.get_legend_handles_labels()
 
@@ -362,3 +364,119 @@ def plot_split_timeseries(
     plt.tight_layout()
 
     return return_axes_data
+
+
+def plot_timeseries_histograms(
+    axes: Axes,
+    data: pd.DataFrame,
+    bins: Union[str, int, np.ndarray, Sequence[Union[int, float]]] = "auto",
+    colormap: Colormap = cm.Blues,
+    **plot_kwargs
+) -> Axes:  # pragma: no cover
+    """Generate a heat-map-like plot for time-series sample data.
+
+    The kind of input this function expects can be obtained from an
+    XArray object as follows:
+
+    .. code:
+
+        data = az_post_trace.posterior_predictive.Y_t[chain_idx].loc[
+            {"dt": slice(t1, t2)}
+        ]
+        data = data.to_dataframe().Y_t.unstack(level=0)
+
+    Parameters
+    ==========
+    axes
+        The Matplotlib axes to use for plotting.
+    data
+        The sample data to be plotted.  This should be in "wide" format: i.e.
+        the index should be "time" and the columns should correspond to each
+        sample.
+    bins
+        The `bins` parameter passed to ``np.histogram``.
+    colormap
+        The Matplotlib colormap use to show relative frequencies within bins.
+    plot_kwargs
+        Keywords passed to ``fill_between``.
+
+    """
+    index = data.index
+    y_samples = data.values
+
+    n_t = len(index)
+
+    # generate histograms and bins
+    list_of_hist, list_of_bins = [], []
+    for t in range(n_t):
+        # TODO: determine proper range=(np.min(Y_t), np.max(Y_t))
+        hist, bins_ = np.histogram(y_samples[t], bins=bins, density=True)
+        if np.sum(hist > 0) == 1:
+            hist, bins_ = np.array([1.0]), np.array([bins_[0], bins_[-1]])
+        list_of_hist.append(hist)
+        list_of_bins.append(bins_)
+
+    if axes is None:
+        _, (axes) = plt.subplots(nrows=1, ncols=1, sharex=True, figsize=(12, 4))
+        axes.plot(index, np.mean(y_samples, axis=1), alpha=0.0, drawstyle="steps")
+
+    for t in range(n_t):
+        mask = index == index[t]
+        hist, bins_ = list_of_hist[t], list_of_bins[t]
+        # normalize bin weights for plotting
+        hist = hist / np.max(hist) * 0.85 if len(hist) > 1 else hist
+        n = len(hist)
+        # construct predictive arrays to plot
+        y_t_ = np.tile(bins_, (n_t, 1))
+        # include consecutive time points to create grid-ish steps
+        if t > 0:
+            mask = np.logical_or(mask, index == index[t - 1])
+        for i in range(n):
+            color_val = hist[i]
+            color = colormap(color_val) if color_val else (1, 1, 1, 1)
+            plot_kwargs.setdefault("step", "pre")
+            axes.fill_between(
+                index,
+                y_t_[:, i],
+                y_t_[:, i + 1],
+                where=mask,
+                color=color,
+                **plot_kwargs,
+            )
+
+    return axes
+
+
+def plot_split_timeseries_histograms(
+    plot_data: pd.DataFrame, sample_col: str, plot_fn: Callable, **split_ts_kwargs
+):  # pragma: no cover
+    """A wrapper function for `plot_split_timeseries` and `plot_timeseries_histograms`
+
+    Parameters
+    ==========
+    plot_data: DataFrame
+        The sample data to be plotted.  This should be in "wide" format: i.e.
+        the index should be "time" and the columns should correspond to each
+        sample.
+    sample_col: str
+        Sample column to be plotted via `plot_timeseries_histograms`.
+    plot_fn: callable
+        A user-defined function to plot non-sample column(s), juxtaposed with
+        the histogram plot of the sample column.
+    split_ts_kwargs
+        Keywords passed to ``plot_split_timeseries``.
+
+    """
+    axes_split_data = plot_split_timeseries(
+        plot_data,
+        plot_fn=plot_fn,
+        **split_ts_kwargs,
+    )
+
+    _ = plot_split_timeseries(
+        plot_data[sample_col],
+        axes_split_data=axes_split_data,
+        plot_fn=plot_timeseries_histograms,
+    )
+
+    return axes_split_data
